#include <stdlib.h>
#include <stdint.h>
#include <Time.h>
#include <ESP8266WiFi.h>

#include "wlan_comm.h"
#include "ds18b20.h"
#include "rtc_eeprom.h"
#include "debug.h"
#include "globals.h"
#include "esp_eeprom.h"
#include "sensor.h"

#define REC_BUF_SIZE 1024
uint8_t rec_buf[REC_BUF_SIZE];    // der zentrale Buffer zur Kommunikation mit dem SocketServer

time_t newMinute;   // die aktuelle Minute - zur Minuten-Wechsel-Erkennung
time_t newHour;     // die aktuelle Stunde - zur Stunden-Wechsel-Erkennung
time_t blockSync;   // die Uhrzeit, ab der wieder mit dem SocketServer gesync't werden darf
time_t nextSync;    // die nächste Sync-Zeit bei fehlgeschlagenem Sync oder 0, wenn gemäß Sync-Zeiten-Tabelle gesync't werden soll
time_t pump_off;    // die Uhrzeit, ab der die Z-Pumpe unbedingt wieder abgeschaltet werden soll oder 0, wenn keine Ausschaltzeit gilt
time_t pump_block;  // die Uhrzeit, ab der die Z-Pumpe frühestens wieder eingeschaltet werden darf

int leg_mode;       // auf 1, wenn z.Zt. der Legionellen-Modus läuft
int leg_time;       // die Uhrzeit, ab der der Legionellen-Modus frühestens abgeschaltet werden darf
int leg_date;       // das Datum, ab dem wieder der Legionellen-Modus aktiviert werden kann
int pump_on;        // auf 1, wenn die Z-Pumpe läuft bzw. laufen soll
int led_rt;         // auf 1, wenn die rote LED leuchtet bzw. leuchten soll
int led_bl;         // auf 1, wenn die blaue LED leuchtet bzw. leuchten soll

/* -------------------------------------------------------------------
 *  Setzt die interne Uhr gemäß des Timestamp-Strings "datetime_str".
 */
void setSysTimeFromString(uint8_t *datetime_str) {
  // Einzelstrings aus Timestamp bauen
  // 2015.10.20 19:48:08
  // 01234567890123456789
  //     X  X  X  X  X
  *(datetime_str+4)='\0';
  *(datetime_str+7)='\0';
  *(datetime_str+10)='\0';
  *(datetime_str+13)='\0';
  *(datetime_str+16)='\0';
  setTime(atoi((char*)datetime_str+11), atoi((char*)datetime_str+14), atoi((char*)datetime_str+17),
          atoi((char*)datetime_str+8), atoi((char*)datetime_str+5), atoi((char*)datetime_str));
}


/* -------------------------------------------------------------------
 *  Init
 *    Sensor-Adressen lesen
 *    loop
 *      Zeit holen und als Systemzeit einstellen
 *      Sensor-Adressen senden
 *      Sensor-Zuordnung empfangen
 *      
 */
void setup() {
  int i, rc, new_data_received;
  time_t t;

  error_g=0x01;
  WiFi.forceSleepBegin();
  #if DEBUG==1
  Serial.begin(115200);
  #endif

  pinMode(RELAIS_PIN, OUTPUT);
  digitalWrite(RELAIS_PIN, LOW);  // Relais aus
  RTC_EEPROM_init();
  pinMode(LED_BL_PIN, OUTPUT);
  pinMode(LED_RT_PIN, OUTPUT);
  digitalWrite(LED_BL_PIN, HIGH);  // blaue LED aus
  digitalWrite(LED_RT_PIN, HIGH);  // rote LED aus

  pinMode(DS18B20_POWER_PIN, OUTPUT);     // Kaltstart für die ds18b20-Sensoren erzwingen
  digitalWrite(DS18B20_POWER_PIN, LOW);   // Vcc der Sensoren auf Masse ziehen
  delay(1);                               // warten, dass der GPIO sicher auf Masse liegt
  resetDS18B20();                         // noch den DQ-Pin kurz auf Masse legen
  delay(1);                               // lieber auch danach noch etwas warten
  digitalWrite(DS18B20_POWER_PIN, HIGH);  // Vcc wieder auf +3,3V
  delay(1);                               // nochmal warten
  resetDS18B20();                         // und noch einen Reset via DQ-Pin

  clearSensorData_subMinuteArray();
  clearSensorDataArray();

  rc=copyVarsFromEEPROM();
  //rc=false;
  if(!rc) {
    // Wenn die Daten im EEPROM nicht valide sind, wurden sie
    // voraussichtlich noch nie gesetzt. Damit der erste Connect
    // mit dem SocketServer klappen kann, müssen sie dann hier
    // gesetzt werden.
    strcpy(SSID[0], "FreifunkWees01.2 (http://ffw)");
    strcpy(SSID[1], "FreifunkWees01.0 (http://ffw)");
    strcpy(SSID[2], "");
    strcpy(SocketServer_adr[0], "192.168.42.80");
    strcpy(SocketServer_adr[1], "");
    SocketServer_port[0]=2629;
    SocketServer_port[1]=0;
    Sync_time[0][0]=3;
    Sync_time[0][1]=15;
    for(i=1; i<NUMBER_SYNCTIMES; i++) {
      Sync_time[i][0]=255;
      Sync_time[i][1]=255;
    }
    #if DEBUG==1
    Serial.println("\n\nFactoryDefaults eingestellt!");
    #endif
    error_g|=0x02;
  }

  rc=discoverOneWireDevices();
  if(rc==0) {
    #if DEBUG==1
    Serial.println("crc-error in discoverOneWireDevices()");
    #endif
    error_g|=0x04;
  }
  if(sensor_count_g==0) {
    strncpy((char*)ROMcode_g[0], "\x28\xff\x94\xba\x02\x17\x04\xc6", 8);  // Fake-Sensoren fürs Debugging auf WeMos
    strncpy((char*)ROMcode_g[1], "\x28\xff\x9e\x25\x03\x17\x04\xab", 8);  // ohne den ",8" würden 9 Byte (+'\0') kopiert werden
    strncpy((char*)ROMcode_g[2], "\x28\xff\x66\xd9\x02\x17\x03\x78", 8);
    strncpy((char*)ROMcode_g[3], "\x28\xff\x28\xb9\x02\x17\x03\xa3", 8);
    sensor_count_g=4;
    error_g|=0x08;
  }
  #if DEBUG==1
  Serial.print("\n\nsensor_count_g=");  Serial.println(sensor_count_g);
  #endif
  if(sensor_count_g!=MAX_SENSOR_COUNT) {
    #if DEBUG==1
    Serial.println("Falsche Anzahl von Sensoren erkannt!");
    #endif
    error_g|=0x08;
  }
  for(i=0; i<sensor_count_g; i++) {     // Dummy-read zwecks Init aller erkannten Sensoren
    rc=readDS18B20_CRC(i);
    if(rc==0x8FFF)  {
      error_g|=0x10;
    }
  }

  doFullSync(true);

  t=now();
  newMinute=t;          // zur Minuten-Wechsel-Erkennung
  newHour=t;            // zur Stunden-Wechsel-Erkennung
  blockSync=0;          // keine Sync-Blockade
  nextSync=0;           // sofortiger Sync
  pump_on=0;            // Pumpe ist aus
  pump_off=0;           // keine Pumpen-Ausschaltzeit
  pump_block=0;         // keine Einschalt-Sperre
  nextSensorHour(t);    // Daten-Speicherung für aktuelle Stunde initialisieren
  led_rt=led_bl=0;      // LEDs aus
  leg_mode=0;           // kein Legionellen-Modus
  leg_time=0;           // keine Endzeit für den Legionellen-Modus
  leg_date=0;           // keine Tages-Blockade für den Legionellen-Modus
}


/* -------------------------------------------------------------------
 *  Liefert true, wenn eine Verbindung zum SocketServer steht.
 */
int reconnectIfNeeded(void) {
  if(isConnectedToSocketServer()) {           // wenn die Verbindung zum SocketServer steht
    return(true);                             // ist alles gut
  }
  if(WiFi.status()!=WL_CONNECTED) {           // wenn WLAN-Connect nicht mehr steht
    if(connectKnownWLAN()) {                  // im WLAN anmelden und wenns geklappt hat
      return(connectKnownSocketServer());     // beim SocketServer anmelden und Erfolgsstatus liefern
    }
  } else {                                    // wenn WLAN-Connect steht
    return(connectKnownSocketServer());       // beim SocketServer anmelden und Erfolgsstatus liefern
  }
  return(false);
}

/* -------------------------------------------------------------------
 *  Fragt beim SocketServer die aktuelle Zeit an und stellt diese
 *  in der RTC und als SystemZeit ein. Liefert true, wenn die Zeit
 *  eingestellt wurde.
 */
int getSetDateTime(void) {
  time_t t;

  sendSocketServer((uint8_t*)"GET_TIME", 8);  // Zeit anfragen
  if(receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000)>=19) {  // Zeit empfangen z.B.: 2017.09.21 20:17:06
    setRTC((char*)rec_buf);
    setSysTimeFromString(rec_buf);
    t=getRTC();
    if(abs(now()-t)>1) {  // bei mehr als einer Sekunde Abweichnung ist was faul
      #if DEBUG==1
      Serial.println("SYS != RTC !!!");
      #endif
      error_g|=0x20;
      return(false);
    }
  } else {
    return(false);
  }
  return(true);
}

/* -------------------------------------------------------------------
 *  
 */
int doFullSync(int fromSetup) {
  int rc, new_data_received=false;
  time_t t;

  digitalWrite(LED_BL_PIN, LOW);  // blaue LED an
  if(reconnectIfNeeded()) {
    getSetDateTime();
    new_data_received|=syncConnectionData();    // WLAN-SSID, IP-Adresse, Portnummer
    if(fromSetup) {                             // nur, wenn aus setup() aufgerufen...
      new_data_received|=syncSensorMetaData();  // ...Sensor-Adress-Zuordnung
    }
    new_data_received|=getPowerOnBitmap();      // Einschaltzeiten
    new_data_received|=getSyncTimes();          // Sync-Zeiten
    new_data_received|=getThresholds();         // Schwellwerte / Deltas
    sendError();                                // Inhalt von error_g senden
    if(!fromSetup) {                            // nur, wenn aus loop() aufgerufen...
      sendSensorHours();                        // ...Sensor-Daten senden
    }
    disconnectSocketServer();
    disconnectWLAN();
    if(new_data_received) {  // wenn Änderungen empfangen wurden...
      rc=copyVarsToEEPROM();
      #if DEBUG==1
      Serial.print("copyVarsToEEPROM()=");      Serial.println(rc);
      #endif
    }
    rc=true;
  } else {
    // die Kommunikation mit dem SocketServer hat nicht geklappt
    t=getRTC();
    setTime(hour(t), minute(t), second(t), day(t), month(t), year(t));
    #if DEBUG==1
    Serial.println("Der Sync war nicht erfolgreich!");
    Serial.print("Die SystemZeit wurde gemäß RTC eingestellt:");
    printRTC(t);  Serial.println();
    #endif
    rc=false;
  }
  digitalWrite(LED_BL_PIN, HIGH);  // blaue LED aus
  return(rc);
}

/* -------------------------------------------------------------------
 *  Sendet den Inhalt der Variablen
 *    SSID[][]
 *    SocketServer_adr[][]
 *    SocketServer_port[]
 *  an den SocketServer.
 *  Bei der Antwort "OK" passiert nix weiter, bei "NEW,..." wird
 *  die entsprechende Variable (im RAM) auf den übermittelten Wert 
 *  gesetzt.
 *  Liefert true, wenn mindestens eine Variable neu gesetzt wurde.
 */
int syncConnectionData(void) {
  int nr, was_changed;

  was_changed=false;
  for(nr=0; nr<3; nr++) {
    strcpy((char*)rec_buf, "SSID.");
    snprintf((char*)rec_buf+strlen((char*)rec_buf), 34, "%d=%s", nr, SSID[nr]);
    sendSocketServer(rec_buf, strlen((char*)rec_buf));
    receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000);
    #if DEBUG==1
    Serial.print("syncSSIDs "); Serial.println((char*)rec_buf);
    #endif
    if(strcmp((char*)rec_buf, "OK")!=0) {
      was_changed=true;
      strncpy(SSID[nr], (char*)rec_buf+4, 33);
    }
  }
  for(nr=0; nr<2; nr++) {
    strcpy((char*)rec_buf, "SOCKSRV_IPADR.");
    snprintf((char*)rec_buf+strlen((char*)rec_buf), 34, "%d=%s", nr, SocketServer_adr[nr]);
    sendSocketServer(rec_buf, strlen((char*)rec_buf));
    receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000);
    #if DEBUG==1
    Serial.print("syncSOCKSRV_IPADRs "); Serial.println((char*)rec_buf);
    #endif
    if(strcmp((char*)rec_buf, "OK")!=0) {
      was_changed=true;
      strncpy(SocketServer_adr[nr], (char*)rec_buf+4, 33);
    }
  }
  for(nr=0; nr<2; nr++) {
    strcpy((char*)rec_buf, "SOCKSRV_PORT.");
    snprintf((char*)rec_buf+strlen((char*)rec_buf), 34, "%d=%d", nr, SocketServer_port[nr]);
    sendSocketServer(rec_buf, strlen((char*)rec_buf));
    receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000);
    #if DEBUG==1
    Serial.print("syncSOCKSRV_PORTs "); Serial.println((char*)rec_buf);
    #endif
    if(strcmp((char*)rec_buf, "OK")!=0) {
      was_changed=true;
      SocketServer_port[nr]=atoi((char*)rec_buf+4);
    }
  }
  return(was_changed);
}

/* -------------------------------------------------------------------
 *  Schreibt die Adresse des Sensors "sn" als String nach "buf".
 */
void formatDS18b20Address(int sn, char *buf) {
  snprintf(buf, 20, "%02x%02x%02x%02x%02x%02x%02x%02x,",
    ROMcode_g[sn][0], ROMcode_g[sn][1], ROMcode_g[sn][2], ROMcode_g[sn][3],
    ROMcode_g[sn][4], ROMcode_g[sn][5], ROMcode_g[sn][6], ROMcode_g[sn][7]);
}

/* -------------------------------------------------------------------
 *  Sendet die Sensor-Adressen und empfängt deren Funktions-Zuordnung.
 *  Liefert true, wenn eine geänderte Sensor-Zuordnung nach
 *  ROMindex_g[] geschrieben wurde.
 */
int syncSensorMetaData(void) {
  int s, ks;

  strcpy((char*)rec_buf, "SENSORS=");
  for(ks=s=0; s<sensor_count_g; s++) {
    formatDS18b20Address(s, (char*)rec_buf+strlen((char*)rec_buf)); // Sensor-Adressen hochmelden
    ks=1;
  }
  if(ks==1) {
    rec_buf[strlen((char*)rec_buf)-1]=0; // das letzte Komma löschen
  }
  sendSocketServer(rec_buf, strlen((char*)rec_buf));
  // Funktions-Zuordnung empfangen
  if(receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000)>0) {
    if(strcmp((char*)rec_buf, "BAD")==0) {
      #if DEBUG==1
      Serial.println("syncSensorMetaData() hat BAD empfangen!");
      #endif
      return(false);
    }
    if(memcmp(rec_buf, ROMindex_g, 4)==0) {
      return(false);  // keine Änderungen empfangen
    }
    for(s=0; s<strlen((char*)rec_buf) && s<MAX_SENSOR_COUNT; s++) {
      ROMindex_g[s]=rec_buf[s];
    }
    #if DEBUG==1
    Serial.println((char*)rec_buf);
    #endif
    return(true);
  }
  return(false);
}


/* -------------------------------------------------------------------
 *  Empfängt die Schaltzeiten-Bitmap für die Zirkulationspumpe. 
 *  Wurden Änderungen erkannt, werden die Daten nach
 *  Power_on_time_g[][] kopiert und true zurückgeliefert.
 */
int getPowerOnBitmap(void) {
  int len;

  strcpy((char*)rec_buf, "GET_POWER_ON_BITMAP");
  sendSocketServer(rec_buf, strlen((char*)rec_buf));
  len=receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000);
  if(len==7*36) {
    if(memcmp(rec_buf, Power_on_time_g, 7*36)==0) {
      return(false);  // keine Änderungen empfangen
    }
    memcpy(Power_on_time_g, rec_buf, 7*36);
    return(true);
  } else {
    #if DEBUG==1
    Serial.print("len(getPowerOnBitmap)=");    Serial.println(len);
    #endif
  }
  return(false);
}

/* -------------------------------------------------------------------
 *  Liefert 1, wenn für Wochtag+Uhrzeit gemäß "t" eine 1 in der
 *  Bitmap steht.
 */
int getPowerOnStateForTime(time_t t) {
  int bit_full_day, byte_adr, bit_num;

  bit_full_day=hour(t)*12+(minute(t)-minute(t)%5)/5;
  byte_adr=bit_full_day/8;
  bit_num=bit_full_day-byte_adr*8;
  //Serial.print("bit_full_day=");    Serial.print(bit_full_day);
  //Serial.print("     byte_adr=");   Serial.print(byte_adr);
  //Serial.print("     bit_num=");    Serial.println(bit_num);
  return((Power_on_time_g[weekday(t)-1][byte_adr]>>bit_num)&1);
}

/* -------------------------------------------------------------------
 *  Empfängt NUMBER_SYNCTIMES Sync-Zeiten.
 *  Wurden Änderungen erkannt, werden die Daten nach
 *  Sync_time[][] kopiert und true zurückgeliefert.
 */
int getSyncTimes(void) {
  int len;

  strcpy((char*)rec_buf, "GET_SYNC_TIMES");
  sendSocketServer(rec_buf, strlen((char*)rec_buf));
  len=receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000);
  if(len==NUMBER_SYNCTIMES*2) {
    if(memcmp(rec_buf, Sync_time, NUMBER_SYNCTIMES*2)==0) {
      return(false);  // keine Änderungen empfangen
    }
    memcpy(Sync_time, rec_buf, NUMBER_SYNCTIMES*2);
    return(true);
  } else {
    #if DEBUG==1
    Serial.print("len(getSyncTimes)=");    Serial.println(len);
    #endif
  }
  return(false);
}

/* -------------------------------------------------------------------
 *  Empfängt die Schwellwerte.
 *  Liefert true, wenn geänderte Schwellwerte nach Threshold[]
 *  geschrieben wurde.
 */
int getThresholds(void) {
  int len;

  strcpy((char*)rec_buf, "GET_THRESHOLDS");
  sendSocketServer(rec_buf, strlen((char*)rec_buf));
  len=receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000);
  if(len==16) {
    if(memcmp(rec_buf, Threshold, 16)==0) {
      return(false);  // keine Änderungen empfangen
    }
    memcpy(Threshold, rec_buf, 16);
    return(true);
  } else {
    #if DEBUG==1
    Serial.print("len(syncThresholds)=");    Serial.println(len);
    #endif
  }
  return(false);
}

/* -------------------------------------------------------------------
 *  Sendet alle validen Daten-Sätze in "sensor_day_g" an den
 *  SocketServer. Gesendete Sätze werden auf invalide gesetzt.
 *  Liefert true, wenn mindestens ein Satz gesendet und bestätigt
 *  wurde.
 */
int sendSensorHours(void) {
  int h, m;
  int len;
  int has_valid_data, ack=false;
  char *ptr;

  for(h=0; h<26; h++) {
    has_valid_data=false;
    for(m=0; m<60; m++) {   // prüfen, ob es mindestens eine gültige Daten-Minute in der Stunde gibt
      if(sensor_day_g[h].data[m].bad==0) {
        has_valid_data=true;
        break;
      }
    }
    if(has_valid_data) {
      snprintf((char*)rec_buf, REC_BUF_SIZE, "SENSORHOUR.%02d=", hour(sensor_day_g[h].timestamp));
      len=strlen((char*)rec_buf);
      memcpy((char*)rec_buf+len, &sensor_day_g[h], sizeof(SensorHour));
      sendSocketServer(rec_buf, len+sizeof(SensorHour));
      receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000);
      if(strcmp((char*)rec_buf, "OK")==0) {
        ack=true;
        clearSensorDataHour(h);
      }
    }
  }
  return(ack);
}

/* -------------------------------------------------------------------
 *  Sendet "error_g" an den SocketServer und setzt "error_g" zurück,
 *  wenn es erfolgreich empfangen wurde.
 */
int sendError(void) {
  snprintf((char*)rec_buf, REC_BUF_SIZE, "ERROR=0x%x,%d", error_g, sensor_count_g);
  sendSocketServer(rec_buf, strlen((char*)rec_buf));
  receiveSocketServer(rec_buf, REC_BUF_SIZE, 2000);
  if(strcmp((char*)rec_buf, "OK")==0) {
    error_g=0;
    return(true);
  }
  return(false);
}



/* -------------------------------------------------------------------
 *  
 */
void loop() {
  static int rc, i, cnt, pps;
  static int sensor_data_valid;
  static int temp[MAX_SENSOR_COUNT];  // E=0, L=1, Z=2, R=3
  static float temp_f[MAX_SENSOR_COUNT];
  static time_t sysTime;

  // ------------------------------------------------------------
  // Zeit pro Durchlauf fix halten
  // ------------------------------------------------------------
  sysTime=now();
  #if DEBUG==1
  Serial.print("SYS:"); printRTC(sysTime);  Serial.println();
  #endif

  // ------------------------------------------------------------
  // Minuten- und Stunden-Umschaltung für die Daten-Speicherung
  // ------------------------------------------------------------
  if(minute(sysTime)!=minute(newMinute)) {
    storeSensorData(newMinute);
    clearSensorData_subMinuteArray();
    #if DEBUG==1
    Serial.println("neue Minute ---------------------------");
    #endif
    newMinute=sysTime;
  }
  if(hour(sysTime)!=hour(newHour)) {
    nextSensorHour(sysTime);
    #if DEBUG==1
    Serial.println("neue Stunde ###########################");
    #endif
    newHour=sysTime;
  }

  // ------------------------------------------------------------
  // Sensoren lesen und gemäß Funktions-Zuordnung speichern
  // ------------------------------------------------------------
  for(i=0; i<sensor_count_g; i++) {
    if(ROMindex_g[i]=='E')      temp[SENSOR_E]=readDS18B20_CRC(i);
    else if(ROMindex_g[i]=='L') temp[SENSOR_L]=readDS18B20_CRC(i);
    else if(ROMindex_g[i]=='Z') temp[SENSOR_Z]=readDS18B20_CRC(i);
    else if(ROMindex_g[i]=='R') temp[SENSOR_R]=readDS18B20_CRC(i);
  }
  for(i=0; i<sensor_count_g; i++) {
    temp_f[i]=temp[i]/16.0;
  }

  // ------------------------------------------------------------
  // auf nicht valide Daten von den Sensoren prüfen
  // ------------------------------------------------------------
  if(temp_f[SENSOR_E]>Threshold[7] || temp_f[SENSOR_L]>Threshold[7] || temp_f[SENSOR_Z]>Threshold[7]) {
    sensor_data_valid=0;
  } else {
    sensor_data_valid=1;
  }

  // ------------------------------------------------------------
  // Z-Pumpe gemäß Zeiten-Bitmap und Temperaturen steuern
  // ------------------------------------------------------------
  if(sensor_data_valid==1) {
    // die SensorDaten erscheinen valide
    // ------------------------------------------------------------
    // Legionellen-Modus
    // ------------------------------------------------------------
    if(leg_mode==0 && (temp_f[SENSOR_E]>Threshold[1] || temp_f[SENSOR_L]>Threshold[2])) {
      // bisher kein Legionellen-Mode aktiv und hohe Temperatur erkannt
      if(sysTime>leg_date && sysTime>pump_block) {
        // Legionellen-Mode war heute noch nicht an und keine Einschaltsperre
        pump_on=1;                          // Z-Pumpe ein
        leg_mode=1;                         // Legionellen-Modus ein
        leg_time=sysTime+Threshold[5]*60;   // in einer Stunde frühestens Z-Pumpe ausschalten
        leg_date=sysTime+60*60*24;          // frühestens morgen wieder Legionellen-Modus aktivieren
        pump_off=leg_time+Threshold[6]*60;  // nur für den Fall, dass die SensorDaten zwischendurch invalide werden
      }
    }
    if(leg_mode==1 && pump_on==1) {
      // der Legionellen-Mode ist aktiv
      if(sysTime>leg_time && temp_f[SENSOR_Z]>Threshold[3]) {
        // Mindesteinschaltzeit überschritten und Temperatur an Sensor(Z) hoch genug
        leg_mode=0;                           // Legionellen-Modus aus
        pump_on=0;                            // Z-Pumpe aus
        pump_block=sysTime+Threshold[10]*60;  // Einschaltsperre Z-Pumpe für die nächsten 5 Minuten
      } else if(sysTime>(leg_time+Threshold[6]*60)) {
        // Maximaleinschaltzeit überschritten
        leg_mode=0;                           // Legionellen-Modus aus
        pump_on=0;                            // Z-Pumpe aus
        pump_block=sysTime+Threshold[10]*60;  // Einschaltsperre Z-Pumpe für die nächsten 5 Minuten
      }
    } else {
      // der Legionellen-Mode ist inaktiv
      // ------------------------------------------------------------
      // Normal-Betrieb
      // ------------------------------------------------------------
      pps=getPowerOnStateForTime(sysTime);
      if(pps==1 && sysTime>pump_block) {
        // wenn die Z-Pumpe laut Zeiten-Bitmap jetzt an sein soll und keine Einschaltsperre
        pump_on=1;                          // Z-Pumpe ein
        pump_off=sysTime+Threshold[4]*60;   // Ausschaltzeit setzen - die 15 Minuten starten erst ab pps==0
      }
      if(pump_on==1) {
        // wenn Pumpe läuft
        if(pump_off>0 && sysTime>pump_off) {
          // wenn Pumpe zusätzliche 15 Minuten an war
          pump_on=0;                            // Z-Pumpe aus
          pump_off=0;                           // Ausschaltzeit zurücksetzen
          pump_block=sysTime+Threshold[10]*60;  // Einschaltsperre Z-Pumpe für die nächsten 5 Minuten
        }
        if(pps==0 && (temp_f[SENSOR_Z]>Threshold[0] || temp_f[SENSOR_Z]+Threshold[9]>temp_f[SENSOR_L])) {
          // wenn Pumpe laut Bitmap nicht laufen muss und es warm genug oder die Speichertemperatur fast erreicht ist
          pump_on=0;                            // Z-Pumpe aus
          pump_off=0;                           // Ausschaltzeit zurücksetzen
          pump_block=sysTime+Threshold[10]*60;  // Einschaltsperre Z-Pumpe für die nächsten 5 Minuten
        }
      }
    }
  } else {    // sensor_data_valid!=1
    // Sensor-Daten sind nicht valide
    // ------------------------------------------------------------
    // Zeitsteuerung (Sensor-Störung)
    // ------------------------------------------------------------
    pps=getPowerOnStateForTime(sysTime);
    if(pps==1 && sysTime>pump_block) {
      // wenn die Z-Pumpe laut Zeiten-Bitmap jetzt an sein soll und keine Einschaltsperre ... wird pump_off neu gesetzt
      pump_on=1;                            // Z-Pumpe an
      leg_mode=0;                           // ggf. Legionellen-Modus aus
      pump_off=sysTime+Threshold[8]*60;     // Pumpe wieder aus, 10 Minuten nach pps==0
    }
    if(pump_on==1 && pump_off>0 && sysTime>pump_off) {
      // wenn Pumpe läuft und die Ausschaltzeit erreicht ist
      pump_on=0;                            // Z-Pumpe aus
      leg_mode=0;                           // ggf. Legionellen-Modus aus
      pump_off=0;                           // Ausschaltzeit zurücksetzen
      pump_block=sysTime+Threshold[10]*60;  // Einschaltsperre Z-Pumpe für die nächsten 5 Minuten
    }
  }

  storeSensorData_subMinute(temp[SENSOR_E], temp[SENSOR_L], temp[SENSOR_Z], temp[SENSOR_R], pump_on);

  // ------------------------------------------------------------
  // Debug
  // ------------------------------------------------------------
  #if DEBUG==1
  Serial.print("Sensoren: E="); Serial.print(temp_f[SENSOR_E]);
  Serial.print("   L=");        Serial.print(temp_f[SENSOR_L]);
  Serial.print("   Z=");        Serial.print(temp_f[SENSOR_Z]);
  Serial.print("   R=");        Serial.println(temp_f[SENSOR_R]);
  #endif

  // ------------------------------------------------------------
  // Sync-Steuerung
  // ------------------------------------------------------------
  for(i=0; i<NUMBER_SYNCTIMES; i++) {
    if(hour(sysTime)==Sync_time[i][0] && minute(sysTime)==Sync_time[i][1] && sysTime>blockSync) {
      // wenn jetzt SyncZeit ist und in dieser Minute noch nicht hierher verzweigt wurde
      nextSync=sysTime;
      blockSync=sysTime+60;  // für den Rest dieser Minute hier das Setzen von nextSync blocken
      #if DEBUG==1
      Serial.print("nextSync=");  printTimestamp(nextSync);  Serial.println();
      #endif
      break;
    }
  }
  if(nextSync>0 && sysTime>=nextSync) {
    if(doFullSync(false)) {
      nextSync=0;     // Sync war erfolgreich
    } else {
      nextSync+=5*60; // Sync war nicht erfolgreich - in fünf Minuten erneut probieren
    }
  }

  // ------------------------------------------------------------
  // Debug
  // ------------------------------------------------------------
  #if DEBUG==1
  Serial.print("SyncTimes=[");
  for(i=0; i<NUMBER_SYNCTIMES; i++) {
    Serial.print("(");
    Serial.print(Sync_time[i][0], DEC);
    Serial.print(", ");
    Serial.print(Sync_time[i][1], DEC);
    Serial.print("), ");
  }
  Serial.println(" ]");
  Serial.print("nextSync=");  printTimestamp(nextSync);  Serial.println();
  Serial.println();
  #endif

  // ------------------------------------------------------------
  // Relais Ansteuerung
  // ------------------------------------------------------------
  if(pump_on==1) {
    digitalWrite(RELAIS_PIN, HIGH);   // Relais ein
  } else {
    digitalWrite(RELAIS_PIN, LOW);    // Relais aus
  }

  // ------------------------------------------------------------
  // LED Ansteuerung
  // ------------------------------------------------------------
  cnt=0;
  led_rt=led_bl=HIGH;
  while(now()<sysTime+5) {
    if((pump_on==1 && cnt%10==0) || (pump_on==0 && cnt%50==0)) {
      led_bl^=1;
    }
    if((error_g>0 && cnt%5==0)) {
      led_rt^=1;
    }
    digitalWrite(LED_RT_PIN, led_rt);
    digitalWrite(LED_BL_PIN, led_bl);
    led_bl=HIGH;  // die blaue LED immer nur 50ms lang leuchten lassen
    cnt++;
    delay(50);    // 5x = 1/4 Sek., 20x = 1 Sek.
  }

}

